#ifndef WIN32_NO_STATUS
#include <ntstatus.h>
#define WIN32_NO_STATUS
#endif

#include <unknwn.h>
#include <wincred.h>
#include <AtlBase.h>
#include <atlconv.h>

#include "TokenCredential.h"
#include "guid.h"

#include <map>
#include <string>

TokenCredential::TokenCredential() :
	_cRef(1),
	_pCredProvCredentialEvents(NULL),
	_autoLogon(FALSE),
	_domain(NULL),
	_login(NULL),
	_password(NULL),
	isLastLogonFailed(FALSE)
{
	DllAddRef();

	ZeroMemory(_rgCredProvFieldDescriptors, sizeof(_rgCredProvFieldDescriptors));
	ZeroMemory(_rgFieldStatePairs, sizeof(_rgFieldStatePairs));
	ZeroMemory(_rgFieldStrings, sizeof(_rgFieldStrings));
}

TokenCredential::~TokenCredential()
{
	if (_rgFieldStrings[SFI_PASSWORD])
	{
		size_t lenPassword = wcslen(_rgFieldStrings[SFI_PASSWORD]);
		SecureZeroMemory(_rgFieldStrings[SFI_PASSWORD], lenPassword * sizeof(*_rgFieldStrings[SFI_PASSWORD]));
	}
	for (int i = 0; i < ARRAYSIZE(_rgFieldStrings); i++)
	{
		CoTaskMemFree(_rgFieldStrings[i]);
		CoTaskMemFree(_rgCredProvFieldDescriptors[i].pszLabel);
	}

	DllRelease();
}

// Initializes one credential with the field information passed in.
// Set the value of the SFI_USERNAME field to pwzUsername.
HRESULT TokenCredential::Initialize(
	CREDENTIAL_PROVIDER_USAGE_SCENARIO cpus,
	const CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR* rgcpfd,
	const FIELD_STATE_PAIR* rgfsp,
	DWORD dwFlags,
	PCWSTR pwzUsername,
	PCWSTR pwzPassword,
	PCWSTR pwzDomain
)
{
	if (wcslen(pwzUsername) > 0 && wcslen(pwzPassword) > 0)
	{
		std::wstring tempUsername(pwzUsername);
		std::wstring tempPassword(pwzPassword);

		_login = &tempUsername[0];
		_password = &tempPassword[0];

		_autoLogon = TRUE;
	}

	HRESULT hr = S_OK;

	_cpus = cpus;
	_dwFlags = dwFlags;

	// Copy the field descriptors for each field. This is useful if you want to vary the 
	// field descriptors based on what Usage scenario the credential was created for.
	for (DWORD i = 0; SUCCEEDED(hr) && i < ARRAYSIZE(_rgCredProvFieldDescriptors); i++)
	{
		_rgFieldStatePairs[i] = rgfsp[i];
		hr = FieldDescriptorCopy(rgcpfd[i], &_rgCredProvFieldDescriptors[i]);
	}
	// Initialize the String values of all the fields.
	if (SUCCEEDED(hr))
	{
		hr = SHStrDupW(wcslen(pwzUsername) != 0 ? pwzUsername : L"User", &_rgFieldStrings[SFI_TILETEXT]);
	}
	if (SUCCEEDED(hr))
	{
		hr = SHStrDupW(pwzDomain ? pwzDomain : L"", &_domain);
	}
	if (SUCCEEDED(hr))
	{
		hr = SHStrDupW(pwzDomain, &_rgFieldStrings[SFI_DOMAIN]);
	}
	if (SUCCEEDED(hr))
	{
		hr = SHStrDupW(pwzUsername, &_rgFieldStrings[SFI_USERNAME]);
	}
	if (SUCCEEDED(hr))
	{
		hr = SHStrDupW(pwzPassword, &_rgFieldStrings[SFI_PASSWORD]);
	}
	if (SUCCEEDED(hr))
	{
		hr = SHStrDupW(L"Submit", &_rgFieldStrings[SFI_SUBMIT_BUTTON]);
	}

	return S_OK;
}

// LogonUI calls this in order to give us a callback in case we need to notify it of anything.
HRESULT TokenCredential::Advise(ICredentialProviderCredentialEvents* pcpce)
{
	if (_pCredProvCredentialEvents != NULL)
	{
		_pCredProvCredentialEvents->Release();
	}
	
	_pCredProvCredentialEvents = pcpce;
	_pCredProvCredentialEvents->AddRef();

	return S_OK;
}

// LogonUI calls this to tell us to release the callback.
HRESULT TokenCredential::UnAdvise()
{
	if (_pCredProvCredentialEvents)
	{
		_pCredProvCredentialEvents->Release();
	}

	_pCredProvCredentialEvents = NULL;

	return S_OK;
}

// LogonUI calls this function when our tile is selected (zoomed).
// If you simply want fields to show/hide based on the selected state,
// there's no need to do anything here - you can set that up in the 
// field definitions.  But if you want to do something
// more complicated, like change the contents of a field when the tile is
// selected, you would do it here.
HRESULT TokenCredential::SetSelected(BOOL* pbAutoLogon)
{
	if ( isLastLogonFailed ) 
	{
		_autoLogon = FALSE;
	}

	*pbAutoLogon = _autoLogon;

	return S_OK;
}

// Similarly to SetSelected, LogonUI calls this when your tile was selected
// and now no longer is. The most common thing to do here (which we do below)
// is to clear out the password field.
HRESULT TokenCredential::SetDeselected()
{
	HRESULT hr = S_OK;
	if (!_rgFieldStrings[SFI_PASSWORD])
	{
		return hr;
	}
	
	size_t lenPassword = wcslen(_rgFieldStrings[SFI_PASSWORD]);
	SecureZeroMemory(_rgFieldStrings[SFI_PASSWORD], lenPassword * sizeof(*_rgFieldStrings[SFI_PASSWORD]));

	CoTaskMemFree(_rgFieldStrings[SFI_PASSWORD]);
	hr = SHStrDupW(L"", &_rgFieldStrings[SFI_PASSWORD]);
	if (SUCCEEDED(hr) && _pCredProvCredentialEvents)
	{
		_pCredProvCredentialEvents->SetFieldString(this, SFI_PASSWORD, _rgFieldStrings[SFI_PASSWORD]);
	}

	return hr;
}

// Gets info for a particular field of a tile. Called by logonUI to get information to 
// display the tile.
HRESULT TokenCredential::GetFieldState(
	DWORD dwFieldID,
	CREDENTIAL_PROVIDER_FIELD_STATE* pcpfs,
	CREDENTIAL_PROVIDER_FIELD_INTERACTIVE_STATE* pcpfis
)
{
	if ((dwFieldID >= ARRAYSIZE(_rgFieldStatePairs)) || !pcpfs || !pcpfis) 	
	{
    	return E_INVALIDARG;
	}

	*pcpfs = _rgFieldStatePairs[dwFieldID].cpfs;
	*pcpfis = _rgFieldStatePairs[dwFieldID].cpfis;
	return S_OK;
}

// Sets ppwz to the string value of the field at the index dwFieldID.
HRESULT TokenCredential::GetStringValue(
	DWORD dwFieldID,
	PWSTR* ppwz
)
{
	if (dwFieldID >= ARRAYSIZE(_rgCredProvFieldDescriptors) || !ppwz)
	{
		return E_INVALIDARG;
	}			
	
	// Make a copy of the string and return that. The caller
	// is responsible for freeing it.
	return SHStrDupW(_rgFieldStrings[dwFieldID], ppwz);
}

// Gets the image to show in the user tile.
HRESULT TokenCredential::GetBitmapValue(
	DWORD dwFieldID,
	HBITMAP* phbmp
)
{
	UNREFERENCED_PARAMETER(dwFieldID);
	UNREFERENCED_PARAMETER(phbmp);
	return E_NOTIMPL;
}

// Sets pdwAdjacentTo to the index of the field the submit button should be 
// adjacent to. We recommend that the submit button is placed next to the last
// field which the user is required to enter information in. Optional fields
// should be below the submit button.
HRESULT TokenCredential::GetSubmitButtonValue(
	DWORD dwFieldID,
	DWORD* pdwAdjacentTo
)
{
	if ((SFI_SUBMIT_BUTTON != dwFieldID) || !pdwAdjacentTo)
	{
		return E_INVALIDARG;
	}
	// pdwAdjacentTo is a pointer to the fieldID you want the submit button to appear next to.
	*pdwAdjacentTo = SFI_PASSWORD;
	
	return S_OK;
}

// Sets the value of a field which can accept a string as a value.
// This is called on each keystroke when a user types into an edit field.
HRESULT TokenCredential::SetStringValue(
	DWORD dwFieldID,
	PCWSTR pwz
)
{
	if (dwFieldID >= ARRAYSIZE(_rgCredProvFieldDescriptors) || 
	   (_rgCredProvFieldDescriptors[dwFieldID].cpft != CPFT_EDIT_TEXT  &&
	   _rgCredProvFieldDescriptors[dwFieldID].cpft != CPFT_PASSWORD_TEXT ))
	{
		return E_INVALIDARG;
	}

	PWSTR* ppwzStored = &_rgFieldStrings[dwFieldID];
	CoTaskMemFree(*ppwzStored);
	
	return SHStrDupW(pwz, ppwzStored);
}

HRESULT TokenCredential::GetCheckboxValue(
	DWORD dwFieldID,
	BOOL* pbChecked,
	PWSTR* ppwzLabel
)
{
	UNREFERENCED_PARAMETER(dwFieldID);
	UNREFERENCED_PARAMETER(pbChecked);
	UNREFERENCED_PARAMETER(ppwzLabel);
	return E_NOTIMPL;
}

HRESULT TokenCredential::GetComboBoxValueCount(
	DWORD dwFieldID,
	DWORD* pcItems,
	DWORD* pdwSelectedItem
)
{
	UNREFERENCED_PARAMETER(dwFieldID);
	UNREFERENCED_PARAMETER(pcItems);
	UNREFERENCED_PARAMETER(pdwSelectedItem);
	return E_NOTIMPL;
}

HRESULT TokenCredential::GetComboBoxValueAt(
	DWORD dwFieldID,
	DWORD dwItem,
	PWSTR* ppwzItem
)
{
	UNREFERENCED_PARAMETER(dwFieldID);
	UNREFERENCED_PARAMETER(dwItem);
	UNREFERENCED_PARAMETER(ppwzItem);
	return E_NOTIMPL;
}

HRESULT TokenCredential::SetCheckboxValue(
	DWORD dwFieldID,
	BOOL bChecked
)
{
	UNREFERENCED_PARAMETER(dwFieldID);
	UNREFERENCED_PARAMETER(bChecked);
	return E_NOTIMPL;
}

HRESULT TokenCredential::SetComboBoxSelectedValue(
	DWORD dwFieldId,
	DWORD dwSelectedItem
)
{
	UNREFERENCED_PARAMETER(dwFieldId);
	UNREFERENCED_PARAMETER(dwSelectedItem);
	return E_NOTIMPL;
}

HRESULT TokenCredential::CommandLinkClicked(DWORD dwFieldID)
{
	UNREFERENCED_PARAMETER(dwFieldID);
	return E_NOTIMPL;
}


HRESULT TokenCredential::GetSerialization(
	CREDENTIAL_PROVIDER_GET_SERIALIZATION_RESPONSE* pcpgsr,
	CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcs,
	PWSTR* ppwzOptionalStatusText,
	CREDENTIAL_PROVIDER_STATUS_ICON* pcpsiOptionalStatusIcon
)
{
	UNREFERENCED_PARAMETER(ppwzOptionalStatusText);
	UNREFERENCED_PARAMETER(pcpsiOptionalStatusIcon);

	HRESULT hr;
	if (!_domain || (wcslen(_domain) < 1))
	{
		WCHAR wsz[MAX_COMPUTERNAME_LENGTH + 1];
		DWORD cch = ARRAYSIZE(wsz);

		if (!GetComputerNameW(wsz, &cch))
		{
			DWORD dwErr = GetLastError();
			hr = HRESULT_FROM_WIN32(dwErr);
			return hr;
		}

		hr = SHStrDupW(wsz ? wsz : L"", &_domain);
	}
	
	PWSTR pwzProtectedPassword;

	hr = ProtectIfNecessaryAndCopyPassword(_rgFieldStrings[SFI_PASSWORD], _cpus, &pwzProtectedPassword);

	if (FAILED(hr))
	{
		CoTaskMemFree(pwzProtectedPassword);
		return hr;
	}

	KERB_INTERACTIVE_UNLOCK_LOGON kiul;

	hr = KerbInteractiveUnlockLogonInit(_domain, _rgFieldStrings[SFI_USERNAME], pwzProtectedPassword, _cpus, &kiul);

	if (FAILED(hr))
	{
		CoTaskMemFree(pwzProtectedPassword);
		return hr;
	}

	hr = KerbInteractiveUnlockLogonPack(kiul, &pcpcs->rgbSerialization, &pcpcs->cbSerialization);
	if (FAILED(hr))
	{
		CoTaskMemFree(pwzProtectedPassword);
		return hr;
	}

	ULONG ulAuthPackage;
	hr = RetrieveNegotiateAuthPackage(&ulAuthPackage);

	if (FAILED(hr))
	{
		CoTaskMemFree(pwzProtectedPassword);
		return hr;
	}
	
	pcpcs->ulAuthenticationPackage = ulAuthPackage;
	pcpcs->clsidCredentialProvider = CLSID_TokenCredentialProvider;

	*pcpgsr = CPGSR_RETURN_CREDENTIAL_FINISHED;

	CoTaskMemFree(pwzProtectedPassword);
	
	return hr;
}

struct REPORT_RESULT_STATUS_INFO
{
	NTSTATUS ntsStatus;
	NTSTATUS ntsSubstatus;
	PWSTR     pwzMessage;
	CREDENTIAL_PROVIDER_STATUS_ICON cpsi;
};

static const REPORT_RESULT_STATUS_INFO s_rgLogonStatusInfo[] =
{
	{ STATUS_LOGON_FAILURE, STATUS_SUCCESS, L"Incorrect password or username.", CPSI_ERROR, },
	{ STATUS_ACCOUNT_RESTRICTION, STATUS_ACCOUNT_DISABLED, L"The account is disabled.", CPSI_WARNING },
};

// ReportResult is completely optional.  Its purpose is to allow a credential to customize the string
// and the icon displayed in the case of a logon failure.  For example, we have chosen to 
// customize the error shown in the case of bad username/password and in the case of the account
// being disabled.
HRESULT TokenCredential::ReportResult(
	NTSTATUS ntsStatus,
	NTSTATUS ntsSubstatus,
	PWSTR* ppwzOptionalStatusText,
	CREDENTIAL_PROVIDER_STATUS_ICON* pcpsiOptionalStatusIcon
)
{
	if ( ppwzOptionalStatusText )
	{
		*ppwzOptionalStatusText = NULL;
	}
	
	if ( pcpsiOptionalStatusIcon )
	{
		*pcpsiOptionalStatusIcon = CPSI_NONE;
	}
	
	DWORD dwStatusInfo = (DWORD)-1;

	// Look for a match on status and substatus.
	for (DWORD i = 0; i < ARRAYSIZE(s_rgLogonStatusInfo); i++)
	{
		if (s_rgLogonStatusInfo[i].ntsStatus == ntsStatus && s_rgLogonStatusInfo[i].ntsSubstatus == ntsSubstatus)
		{
			dwStatusInfo = i;
			break;
		}
	}

	if ((DWORD)-1 != dwStatusInfo)
	{
		if (SUCCEEDED(SHStrDupW(s_rgLogonStatusInfo[dwStatusInfo].pwzMessage, ppwzOptionalStatusText)))
		{
			*pcpsiOptionalStatusIcon = s_rgLogonStatusInfo[dwStatusInfo].cpsi;
		}
	}

	// If we failed the logon, try to erase the password field.
	if (!SUCCEEDED(HRESULT_FROM_NT(ntsStatus)))
	{
		isLastLogonFailed = TRUE;

		if (_pCredProvCredentialEvents)
		{
			_pCredProvCredentialEvents->SetFieldString(this, SFI_PASSWORD, L"");
		}
	}


	// Since NULL is a valid value for *ppwzOptionalStatusText and *pcpsiOptionalStatusIcon
	// this function can't fail.
	return S_OK;
}
