#include <credentialprovider.h>
#include "TokenProvider.h"
#include "TokenCredential.h"
#include "guid.h"
#include <wincred.h>


TokenProvider::TokenProvider():
_cRef(1),
_pkiulSetSerialization(NULL),
_dwCredUIFlags(0),
_bRecreateEnumeratedCredentials(true),
_bAutoSubmitSetSerializationCred(false),
_bDefaultToFirstCredential(false)
{
    DllAddRef();
   
    _jsonReply = ConvertJsonToDict(GetTokenFromPipe());

    ZeroMemory(_rgpCredentials, sizeof(_rgpCredentials));
}

TokenProvider::~TokenProvider()
{
    _ReleaseEnumeratedCredentials();
    DllRelease();
}

void TokenProvider::_ReleaseEnumeratedCredentials()
{
    for (int i = 0; i < ARRAYSIZE(_rgpCredentials); i++)
    {
        if (_rgpCredentials[i] != NULL)
        {
            _rgpCredentials[i]->Release();
        }
    }
}


// SetUsageScenario is the provider's cue that it's going to be asked for tiles
// in a subsequent call.
//
// This sample only handles the logon, unlock, and credui scenarios.
HRESULT TokenProvider::SetUsageScenario(
    CREDENTIAL_PROVIDER_USAGE_SCENARIO cpus,
    DWORD dwFlags
    )
{
    HRESULT hr;

    _cpus = cpus;
    if (cpus == CPUS_CREDUI)
    {
        _dwCredUIFlags = dwFlags;  // currently the only flags ever passed in are only valid for the credui scenario
    }
    _bRecreateEnumeratedCredentials = true;

    switch (cpus)
    {
    case CPUS_LOGON:
    case CPUS_UNLOCK_WORKSTATION:
    case CPUS_CREDUI:
        hr = S_OK;
        break;

    case CPUS_CHANGE_PASSWORD:
        hr = E_NOTIMPL;
        break;

    default:
        hr = E_INVALIDARG;
        break;
    }

    return hr;
}

STDMETHODIMP TokenProvider::SetSerialization(
    const CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcs
    )
{
    HRESULT hr = E_INVALIDARG;

    if ((CLSID_TokenCredentialProvider != pcpcs->clsidCredentialProvider) || (CPUS_CREDUI != _cpus))
    {
        return hr;
    }
    // Get the current AuthenticationPackageID that we are supporting
    ULONG ulNegotiateAuthPackage;
    hr = RetrieveNegotiateAuthPackage(&ulNegotiateAuthPackage);

    if (FAILED(hr))
    {
        return hr;
    }

    if (CPUS_CREDUI == _cpus)
    {
        if (CREDUIWIN_IN_CRED_ONLY & _dwCredUIFlags)
        {
            // If we are being told to enumerate only the incoming credential, we must not return
            // success unless we can enumerate it.  We'll set hr to failure here and let it be
            // overridden if the enumeration logic below succeeds.
            hr = E_INVALIDARG;
        }
        else if (_dwCredUIFlags & CREDUIWIN_AUTHPACKAGE_ONLY)
        {
            if (ulNegotiateAuthPackage != pcpcs->ulAuthenticationPackage)
            {
                hr = E_INVALIDARG;
            }
            hr = S_FALSE;
        }
    }

    if ((ulNegotiateAuthPackage != pcpcs->ulAuthenticationPackage) ||
        (pcpcs->cbSerialization <= 0 || !pcpcs->rgbSerialization))
    {
        return hr;
    }

    KERB_INTERACTIVE_UNLOCK_LOGON* pkil = (KERB_INTERACTIVE_UNLOCK_LOGON*) pcpcs->rgbSerialization;
    if (KerbInteractiveLogon != pkil->Logon.MessageType)
    {
        return hr;
    }

    // If there isn't a username, we can't serialize or create a tile for this credential.
    if (pkil->Logon.UserName.Length <= 0 || !pkil->Logon.UserName.Buffer)
    {
        return hr;
    }

    if ((CPUS_CREDUI != _cpus) || (CREDUIWIN_PACK_32_WOW & _dwCredUIFlags))
    {
        BYTE* rgbSerialization;
        rgbSerialization = (BYTE*)HeapAlloc(GetProcessHeap(), 0, pcpcs->cbSerialization);
        HRESULT hrCreateCred = rgbSerialization ? S_OK : E_OUTOFMEMORY;

        if (FAILED(hrCreateCred))
        {
            if (_pkiulSetSerialization)
            {
                HeapFree(GetProcessHeap(), 0, _pkiulSetSerialization);
            }
            return hr;
        }
    
        CopyMemory(rgbSerialization, pcpcs->rgbSerialization, pcpcs->cbSerialization);
        KerbInteractiveUnlockLogonUnpackInPlace((KERB_INTERACTIVE_UNLOCK_LOGON*)rgbSerialization,pcpcs->cbSerialization);

        if (_pkiulSetSerialization)
        {
            HeapFree(GetProcessHeap(), 0, _pkiulSetSerialization);
        }

        _pkiulSetSerialization = (KERB_INTERACTIVE_UNLOCK_LOGON*)rgbSerialization;
        if (FAILED(hrCreateCred))
        {
            return hr;
        }

        // we allow success to override the S_FALSE for the CREDUIWIN_AUTHPACKAGE_ONLY, but
        // failure to create the cred shouldn't override that we can still handle
        // the auth package
        return hrCreateCred;
    }
    
    BYTE* rgbNativeSerialization;
    DWORD cbNativeSerialization;
    if (SUCCEEDED(KerbInteractiveUnlockLogonRepackNative(pcpcs->rgbSerialization, pcpcs->cbSerialization, &rgbNativeSerialization, &cbNativeSerialization)))
    {
        KerbInteractiveUnlockLogonUnpackInPlace((PKERB_INTERACTIVE_UNLOCK_LOGON)rgbNativeSerialization, cbNativeSerialization);

        _pkiulSetSerialization = (PKERB_INTERACTIVE_UNLOCK_LOGON)rgbNativeSerialization;
        hr = S_OK;
    }

    return hr;
}

// Called by LogonUI to give you a callback.  Providers often use the callback if they
// some event would cause them to need to change the set of tiles that they enumerated
HRESULT TokenProvider::Advise(ICredentialProviderEvents* pcpe,
                              UINT_PTR upAdviseContext)
{
    UNREFERENCED_PARAMETER(pcpe);
    UNREFERENCED_PARAMETER(upAdviseContext);
    return E_NOTIMPL;
}

// Called by LogonUI when the ICredentialProviderEvents callback is no longer valid.
HRESULT TokenProvider::UnAdvise()
{
    return E_NOTIMPL;
}

// Called by LogonUI to determine the number of fields in your tiles.  This
// does mean that all your tiles must have the same number of fields.
// This number must include both visible and invisible fields. If you want a tile
// to have different fields from the other tiles you enumerate for a given usage
// scenario you must include them all in this count and then hide/show them as desired 
// using the field descriptors.
HRESULT TokenProvider::GetFieldDescriptorCount(
    DWORD* pdwCount
    )
{
    *pdwCount = SFI_NUM_FIELDS;

    return S_OK;
}

// Gets the field descriptor for a particular field
HRESULT TokenProvider::GetFieldDescriptorAt(
    DWORD dwIndex, 
    CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR** ppcpfd
    )
{    
    if ((dwIndex >= SFI_NUM_FIELDS) || !ppcpfd)
    {
        return E_INVALIDARG;
    }
    
    return FieldDescriptorCoAllocCopy(s_rgCredProvFieldDescriptors[dwIndex], ppcpfd);
}

// Sets pdwCount to the number of tiles that we wish to show at this time.
// Sets pdwDefault to the index of the tile which should be used as the default.
// The default tile is the tile which will be shown in the zoomed view by default. If 
// more than one provider specifies a default tile the behavior is the last cred prov
// get to specify the default tile.
// If *pbAutoLogonWithDefault is TRUE, LogonUI will immediately call GetSerialization
// on the credential you've specified as the default and will submit that credential
// for authentication without showing any further UI.
HRESULT TokenProvider::GetCredentialCount(
    DWORD* pdwCount,
    DWORD* pdwDefault,
    BOOL* pbAutoLogonWithDefault
    )
{
    HRESULT hr = E_FAIL;
    if (_bRecreateEnumeratedCredentials)
    {
        _ReleaseEnumeratedCredentials();
        hr = _CreateEnumeratedCredentials();
        _bRecreateEnumeratedCredentials = false;
    }

    *pdwCount = 0;
    *pdwDefault = (_bDefaultToFirstCredential && _rgpCredentials[0]) ? 0 : CREDENTIAL_PROVIDER_NO_DEFAULT;
    *pbAutoLogonWithDefault = FALSE;

    if (FAILED(hr))
    {
        return hr;
    }

    // TODO: it would probably be nicer to keep a count of the number of creds
    DWORD dwNumCreds = 0;
    for (int i = 0; i < MAX_CREDENTIALS; i++)
    {
        if (_rgpCredentials[i] != NULL)
        {
            dwNumCreds++;
        }
    }

    switch(_cpus)
    {
    case CPUS_LOGON:
        if (_bAutoSubmitSetSerializationCred)
        {
            *pdwCount = 1;
            *pbAutoLogonWithDefault = TRUE;
        }
        else
        {
            *pdwCount = dwNumCreds;
            // since we have more than one tile and don't keep track of who logged on last, we don't really have a default in this case
        }
        hr = S_OK;
        break;

    case CPUS_UNLOCK_WORKSTATION:
        // in the unlock case, you likely would want to only enumerate tiles for the logged on user (that could be used to unlock)
        // but that's a bit complicated for a sample, so we'll just use our normal tiles
        // that we already set up in the logon case.  The default out params set up at the top work for this case.
        *pdwCount = dwNumCreds;
        hr = S_OK;
        break;

    case CPUS_CREDUI:
        *pdwCount = dwNumCreds;
        hr = S_OK;
        break;

    default:
        hr = E_INVALIDARG;
        break;
    }

    return hr;
}

// Returns the credential at the index specified by dwIndex. This function is called by logonUI to enumerate
// the tiles.
HRESULT TokenProvider::GetCredentialAt(
    DWORD dwIndex, 
    ICredentialProviderCredential** ppcpc
    )
{
    HRESULT hr;

    // Validate parameters.
    if((dwIndex < ARRAYSIZE(_rgpCredentials)) && _rgpCredentials[dwIndex] != NULL && ppcpc)
    {
        hr = _rgpCredentials[dwIndex]->QueryInterface(IID_ICredentialProviderCredential, reinterpret_cast<void**>(ppcpc));
    }
    else
    {
        hr = E_INVALIDARG;
    }

    return hr;
}

// Creates a Credential with the SFI_USERNAME field's value set to pwzUsername.
HRESULT TokenProvider::_EnumerateOneCredential(
    DWORD dwCredentialIndex,
    PCWSTR pwzUsername,
    PCWSTR pwzPassword,
    PCWSTR pwzDomain
    )
{
    HRESULT hr;

    TokenCredential* ppc = new TokenCredential();

    if (!ppc)
    {
        return E_OUTOFMEMORY;
    }

    // Set the Field State Pair and Field Descriptors for ppc's fields
    // to the defaults (s_rgCredProvFieldDescriptors, and s_rgFieldStatePairs) and the value of SFI_USERNAME
    // to pwzUsername.
    hr = ppc->Initialize(_cpus,s_rgCredProvFieldDescriptors, s_rgFieldStatePairs, _dwCredUIFlags, pwzUsername, pwzPassword, pwzDomain);

    if (FAILED(hr))
    {
        ppc->Release();
        return hr;   
    }

    _rgpCredentials[dwCredentialIndex] = ppc;

    return hr;
}

//
// Depending on whether SetSerialization has been called and what CPUS we're in
// creates the right set of credentials
//
HRESULT TokenProvider::_CreateEnumeratedCredentials()
{
    HRESULT hr = E_INVALIDARG;
    
    switch(_cpus)
    {
    case CPUS_LOGON:
        hr = _pkiulSetSerialization ? _EnumerateSetSerialization() : _EnumerateCredentials();
        break;

    case CPUS_CHANGE_PASSWORD:
        break;

    case CPUS_UNLOCK_WORKSTATION:
        // a more advanced implementation would only enumerate tiles that could gather creds for the logged on user
        // since those are the only creds that will work to unlock the session
        // but we're going with this for simplicity
        hr = _EnumerateCredentials();  
        break;

    case CPUS_CREDUI:
        _bDefaultToFirstCredential = true;

        if (_pkiulSetSerialization)
        {
            hr = _EnumerateSetSerialization();
        }
        if (_dwCredUIFlags & CREDUIWIN_ENUMERATE_ADMINS)
        {
            // this sample doesn't handle this particular case
            // You would want to handle this in order to participate
            // User Account Control elevations for a non-admin users (where all the
            // admins tiles are enumerated)
        }
        else if (!(_dwCredUIFlags & CREDUIWIN_IN_CRED_ONLY))
        {
            // if we're here, then we're supposed to enumerate whatever we should enumerate for the normal case.  
            // In our case, that's our 2 tiles.  We may already have one tile, though
            if (_pkiulSetSerialization && SUCCEEDED(hr))
            {
                hr = _EnumerateCredentials(true);
            }
            else
            {
                hr = _EnumerateCredentials(false);
            }
        }
        break;

    default:
        break;
    }
    return hr;
}

HRESULT TokenProvider::_EnumerateCredentials(bool bAlreadyHaveSetSerializationCred)
{
    DWORD dwStart = bAlreadyHaveSetSerializationCred ? 1 : 0;
    
    PWSTR username = L"";
    PWSTR password = L"";
    PWSTR domain = L"";

    if (_jsonReply.count("login") == 1)
    {
        username = s2ws(_jsonReply.at("login"));
    }

    if (_jsonReply.count("pass") == 1)
    {
        password = s2ws(_jsonReply.at("pass"));
    }

    if (_jsonReply.count("domain") == 1)
    {
        domain = s2ws(_jsonReply.at("domain"));        
    }

    // Sets up the tiles for this provider
    return _EnumerateOneCredential(dwStart++, username, password, domain);
}

// This enumerates a tile for the info in _pkiulSetSerialization.  See the SetSerialization function comment for
// more information.
HRESULT TokenProvider::_EnumerateSetSerialization()
{
    KERB_INTERACTIVE_LOGON* pkil = &_pkiulSetSerialization->Logon;

    _bAutoSubmitSetSerializationCred = false;
    _bDefaultToFirstCredential = false;

    WCHAR wszUsername[MAX_PATH] = {0};
    WCHAR wszPassword[MAX_PATH] = {0};

    HRESULT hr = StringCbCopyNW(wszUsername, sizeof(wszUsername), pkil->UserName.Buffer, pkil->UserName.Length);
    if (FAILED(hr))
    {
        return hr;
    }
    
    hr = StringCbCopyNW(wszPassword, sizeof(wszPassword), pkil->Password.Buffer, pkil->Password.Length);
    if (FAILED(hr))
    {
        return hr;
    }

    TokenCredential* pCred = new TokenCredential();
    if (!pCred)
    {
        return E_OUTOFMEMORY;
    }

    hr = pCred->Initialize(_cpus, s_rgCredProvFieldDescriptors, s_rgFieldStatePairs, _dwCredUIFlags, wszUsername, wszPassword);
    if (FAILED(hr))
    {
        return hr;
    }

    _rgpCredentials[0] = pCred;

    _bDefaultToFirstCredential = true;  

    if (SUCCEEDED(hr) && (0 < wcslen(wszPassword)))
    {
        _bAutoSubmitSetSerializationCred = true;
    }

    return hr;
}

HRESULT TokenProvider_CreateInstance(REFIID riid, void** ppv)
{
    HRESULT hr;

    TokenProvider* pProvider = new TokenProvider();

    if (!pProvider)
    {
        return E_OUTOFMEMORY;
    }
    
    hr = pProvider->QueryInterface(riid, ppv);
    pProvider->Release();

    return hr;
}
